
/*
    Tucnak - VHF contest log
    Copyright (C) 2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

struct qrv *qrv;


struct qrv *init_qrv(void){
    struct qrv *qrv;

    qrv = g_new0(struct qrv, 1);
    init_list(qrv->list);
    qrv->cur=g_new0(struct qrv_item *, 'Z'-'A'+1);
        
    return qrv;
}

void free_qrv_item(struct qrv_item *qi){
    CONDGFREE(qi->call);
    CONDGFREE(qi->wwl);
    CONDGFREE(qi->text);
    g_free(qi);
}

void free_qrv(struct qrv *qrv){
    while(qrv->list.next != &qrv->list){
        struct qrv_item *qi;
        qi = qrv->list.next;
        del_from_list(qi);
        free_qrv_item(qi);
    }
    g_free(qrv->cur);
    g_free(qrv);
}


void qrv_add(const gchar *call, const gchar *wwl, int qrv_int, const gchar *text){
    struct qrv_item *qi, *q;
    int ret;
                
    foreach(q, qrv->list) {
        ret = strcasecmp(call, q->call);
        if (ret==0){ // found same, update info
            if (wwl && *wwl) {
                CONDGFREE(q->wwl);
                q->wwl = uc(g_strdup(wwl)); 
            }
            q->bands |= qrv_int;
            if (text && strlen(text) > strlen(q->text)){
                CONDGFREE(q->text);
                q->text = g_strdup(text);
            }
            return;
        }
        if (ret<=0) break; 
    }
    
    qi = g_new0(struct qrv_item, 1);
    qi->call  = uc(g_strdup(call));
    qi->wwl   = uc(g_strdup(wwl));
    qi->bands = qrv_int;
    qi->text  = g_strdup(text);
   
    add_at_pos(q->prev, qi);
    //printf("add_to_list(%s %s %d %s)\r\n", qi->call, qi->wwl, qi->bands, qi->text);
}

#define QRV_DELIM " \t\r\n"
void load_one_qrv(struct qrv *qrv, gchar *s){
    gchar *call, *wwl, *qrv_str, *text;
    int qrv_int, len;
    char *token_ptr, *c;

    call = strtok_r(s, QRV_DELIM, &token_ptr);
    if (!call) return;

    wwl = strtok_r(NULL, QRV_DELIM, &token_ptr);
    if (!wwl) return;

    qrv_str = strtok_r(NULL, QRV_DELIM, &token_ptr);
    if (!qrv_str) {
        qrv_int = -1; // all bands
        text = "";
    }else{
        qrv_int = 0;
        for (c=qrv_str;*c!='\0';c++){
            *c=upcase(*c);
            qrv_int|=1<<(*c-'A');
        }
        
        text = token_ptr; //strtok_r(NULL, QRV_DELIM, &token_ptr);
        if (!text) text = "";
        len=strlen(text);
        if (len>0 && text[len-1]=='\n') text[len-1]='\0';
        len=strlen(text);
        if (len>0 && text[len-1]=='\r') text[len-1]='\0';
    }

    qrv_add(call, wwl, qrv_int, text);
}

int load_qrv_from_file(struct qrv *qrv, gchar *filename){
    FILE *f;
    char s[202];

    f = fopen(filename, "rt");
    if (!f){
        return -1;
    }

    while((fgets(s, 200, f))!=NULL){
        load_one_qrv(qrv,s);
    }
    fclose(f);
    return 0;
}


int save_qrv_files(struct qrv *qrv){
    return 0;
}

int qrv_skip(struct qrv_item *qi, int bi){
    if (get_qso_by_callsign(aband, qi->call)!=NULL) return 1;
    if ((qi->bands & (1<<bi))==0) return 1;
    return 0;
}

void qrv_seek(struct qrv *qrv, int value, int whence){
    int bi;
    struct qrv_item *qi = NULL;
                
    if (!aband) return;
    bi=lowcase(aband->bandchar)-'a';

   // dbg("\nqrv_seek(%d, %d)\n", value, whence);

    if (whence==1) qrv->cur[bi]=qrv->list.next;
    if (whence==-1) qrv->cur[bi]=qrv->list.prev;


    if (value<0) {
        for(qi = qrv->cur[bi]; value < 0; ){
            qi = qi->prev;
            //printf("try %s\n", qi->call);
            if (qi == &qrv->list) return;
            if (qrv_skip(qi, bi)) continue;
            value++;
            qrv->cur[bi] = qi;
        }
        return;
    }
    
    if (value>0) {
        for(qi = qrv->cur[bi]; value > 0; ){
            qi = qi->next;
            //printf("try %s\n", qi->call);
            if (qi == &qrv->list) return;
            if (qrv_skip(qi, bi)) continue;
            value--;
            qrv->cur[bi] = qi;
        }
        return;
    }
}


int sw_qrv_kbd_func(struct subwin *sw, struct event *ev, int fw){
    int bi;
    
    if (!aband) return 0;
    bi=lowcase(aband->bandchar)-'a';
    
    switch(kbd_action(KM_MAIN,ev)){
        case ACT_ESC:
            return 0;
            break;
        case ACT_DOWN:
            qrv_seek(qrv, 1, 0);
            redraw_later();
            return 1;
        case ACT_UP:
            qrv_seek(qrv, -1, 0);
            redraw_later();
            return 1;
        case ACT_PAGE_DOWN:
            qrv_seek(qrv, sw->h-1, 0);
            redraw_later();
            return 1;
        case ACT_PAGE_UP:
            qrv_seek(qrv, -(sw->h-1), 0);
            redraw_later();
            return 1;
        case ACT_HOME:
            qrv_seek(qrv, 1, 1); /* seek to end */
            redraw_later();
            return 1;
        case ACT_END:
            qrv_seek(qrv, 1, -1); /* seek to end */
            redraw_later();
            return 1;
		case ACT_SCROLL_LEFT:
    		if (sw->ho>0) sw->ho--;
            redraw_later();
            return 1;
		case ACT_SCROLL_RIGHT:
			sw->ho++;
            redraw_later();
            return 1;
        case ACT_ENTER:
            sw_unset_focus();
            il_set_focus(aband->il);
            if (qrv->cur[bi]){
                clear_tmpqsos(aband,1);
                process_input(aband, qrv->cur[bi]->call);
                redraw_later();
/*#ifdef HAVE_HAMLIB                
                if (trig){
                    trig_set_qrg(trig, qrv->cur[bi]->qrg*1000.0);
                }
#endif                */
            }
            return 1;
        case ACT_DELETE:
            if (!qrv->cur[bi]) break;
            qrv->cur[bi]->bands &= ~(1<<bi);
            qrv_seek(qrv, 1, 0);
            redraw_later();
            return 1;
        default:
            if (ev->x==' ') {
                if (qrv->cur[bi]){
                    clear_tmpqsos(aband,1);
                    process_input(aband, qrv->cur[bi]->call);
                    redraw_later();
                }
            }
    }
    return 0;
}

int sw_qrv_mouse_func(struct subwin *sw, struct event *ev, int fw){
    return 0;
}

void sw_qrv_draw_qrv(struct subwin *sw, int i, struct qrv_item *qi, int color){
    gchar *c;

    if (!qi || !qi->call){
        c = g_strdup("~");
    }else{
        c = g_strdup_printf("%-12s  %6s  %s", qi->call, qi->wwl, qi->text);
    }

    if (c && strlen(c)>sw->ho){
        print_text(sw->x, sw->y+i, sw->w, c + sw->ho, color);
    }
    g_free(c);
}


void sw_qrv_redraw(struct subwin *sw){
    struct qrv_item *qi;
    int i,bi;

    if (!aband) return;
    
    bi=lowcase(aband->bandchar)-'a';
    fill_area(sw->x, sw->y+sw->h/2, sw->w, 1, COL_INV);
    sw_qrv_draw_qrv(sw, sw->h/2, qrv->cur[bi], COL_INV);
   
    if (!qrv->cur[bi]) qrv_seek(qrv, 1, 1);

    if (qrv->cur[bi]) qi=qrv->cur[bi]->prev;
    else              qi=NULL;
    
    for (i=sw->h/2-1;i>=0; i--){
        if (!qi) break;
        if (qi==&qrv->list) break;
        if (qrv_skip(qi, bi))
            i++;
        else
            sw_qrv_draw_qrv(sw, i, qi, COL_NORM);
        
        qi=qi->prev;
    }
    
    if (qrv->cur[bi]) qi=qrv->cur[bi]->next;
    else              qi=NULL;

    for (i=sw->h/2+1;i<sw->h; i++){
        if (!qi) break;
        if (qi==&qrv->list) break;
        if (qrv_skip(qi, bi))
            i--;
        else
            sw_qrv_draw_qrv(sw, i, qi, COL_NORM);
        qi=qi->next;
    }
}

void sw_qrv_check_bounds(struct subwin *sw){
}
